import math


def in_sine(step_ratio: float) -> float:
    """
    Ease in using a sine function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return 1 - math.cos((step_ratio * math.pi) / 2)


def out_sine(step_ratio: float) -> float:
    """
    Ease out using a sine function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return math.sin((step_ratio * math.pi) / 2)


def in_out_sine(step_ratio: float) -> float:
    """
    Ease in/out using a sine function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """

    return -(math.cos(math.pi * step_ratio) - 1) / 2


def in_quad(step_ratio: float) -> float:
    """
    Ease in using a quadratic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """

    return step_ratio**2


def out_quad(step_ratio: float) -> float:
    """
    Ease out using a quadratic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel

    """
    return 1 - (1 - step_ratio) * (1 - step_ratio)


def in_out_quad(step_ratio: float) -> float:
    """
    Ease in/out using a quadratic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel

    """
    if step_ratio < 0.5:
        return 2 * step_ratio**2
    else:
        return 1 - (-2 * step_ratio + 2) ** 2 / 2


def in_cubic(step_ratio: float) -> float:
    """
    Ease in using a cubic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return step_ratio**3


def out_cubic(step_ratio: float) -> float:
    """
    Ease out using a cubic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return 1 - (1 - step_ratio) ** 3


def in_out_cubic(step_ratio: float) -> float:
    """
    Ease in/out using a cubic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percentage of the current waypoint speed to apply to the
        character
    """
    if step_ratio < 0.5:
        return 4 * step_ratio**3
    else:
        return 1 - (-2 * step_ratio + 2) ** 3 / 2


def in_quart(step_ratio: float) -> float:
    """
    Ease in using a quartic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percentage
        of the current waypoint speed to apply to the character
    """
    return step_ratio**4


def out_quart(step_ratio: float) -> float:
    """
    Ease out using a quartic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return 1 - (1 - step_ratio) ** 4


def in_out_quart(step_ratio: float) -> float:
    """
    Ease in/out using a quartic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio < 0.5:
        return 8 * step_ratio**4
    else:
        return 1 - (-2 * step_ratio + 2) ** 4 / 2


def in_quint(step_ratio: float) -> float:
    """
    Ease in using a quintic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return step_ratio**5


def out_quint(step_ratio: float) -> float:
    """
    Ease out using a quintic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return 1 - (1 - step_ratio) ** 5


def in_out_quint(step_ratio: float) -> float:
    """
    Ease in/out using a quintic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio < 0.5:
        return 16 * step_ratio**5
    else:
        return 1 - (-2 * step_ratio + 2) ** 5 / 2


def in_expo(step_ratio: float) -> float:
    """
    Ease in using an exponential function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio == 0:
        return 0
    else:
        return 2 ** (10 * step_ratio - 10)


def out_expo(step_ratio: float) -> float:
    """
    Ease out using an exponential function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio == 1:
        return 1
    else:
        return 1 - 2 ** (-10 * step_ratio)


def in_out_expo(step_ratio: float) -> float:
    """
    Ease in/out using an exponential function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio == 0:
        return 0
    elif step_ratio == 1:
        return 1
    elif step_ratio < 0.5:
        return 2 ** (20 * step_ratio - 10) / 2
    else:
        return (2 - 2 ** (-20 * step_ratio + 10)) / 2


def in_circ(step_ratio: float) -> float:
    """
    Ease in using a circular function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return 1 - math.sqrt(1 - step_ratio**2)


def out_circ(step_ratio: float) -> float:
    """
    Ease out using a circular function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    return math.sqrt(1 - (step_ratio - 1) ** 2)


def in_out_circ(step_ratio: float) -> float:
    """
    Ease in/out using a circular function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio < 0.5:
        return (1 - math.sqrt(1 - (2 * step_ratio) ** 2)) / 2
    else:
        return (math.sqrt(1 - (-2 * step_ratio + 2) ** 2) + 1) / 2


def in_back(step_ratio: float) -> float:
    """
    Ease in using a back function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    c1 = 1.70158
    c3 = c1 + 1
    return c3 * step_ratio**3 - c1 * step_ratio**2


def out_back(step_ratio: float) -> float:
    """
    Ease out using a back function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    c1 = 1.70158
    c3 = c1 + 1
    return 1 + c3 * (step_ratio - 1) ** 3 + c1 * (step_ratio - 1) ** 2


def in_out_back(step_ratio: float) -> float:
    """
    Ease in/out using a back function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    c1 = 1.70158
    c2 = c1 * 1.525
    if step_ratio < 0.5:
        return ((2 * step_ratio) ** 2 * ((c2 + 1) * 2 * step_ratio - c2)) / 2
    else:
        return ((2 * step_ratio - 2) ** 2 * ((c2 + 1) * (step_ratio * 2 - 2) + c2) + 2) / 2


def in_elastic(step_ratio: float) -> float:
    """
    Ease in using an elastic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """

    c4 = (2 * math.pi) / 3
    if step_ratio == 0:
        return 0
    elif step_ratio == 1:
        return 1
    else:
        return -(2 ** (10 * step_ratio - 10)) * math.sin((step_ratio * 10 - 10.75) * c4)


def out_elastic(step_ratio: float) -> float:
    """
    Ease out using an elastic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percentage of the current waypoint speed to apply to the character
    """
    c4 = (2 * math.pi) / 3
    if step_ratio == 0:
        return 0
    elif step_ratio == 1:
        return 1
    else:
        return 2 ** (-10 * step_ratio) * math.sin((step_ratio * 10 - 0.75) * c4) + 1


def in_out_elastic(step_ratio: float) -> float:
    """
    Ease in/out using an elastic function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percentage of the current waypoint speed to apply to the character
    """
    c5 = (2 * math.pi) / 4.5
    if step_ratio == 0:
        return 0
    elif step_ratio == 1:
        return 1
    elif step_ratio < 0.5:
        return -(2 ** (20 * step_ratio - 10) * math.sin((20 * step_ratio - 11.125) * c5)) / 2
    else:
        return (2 ** (-20 * step_ratio + 10) * math.sin((20 * step_ratio - 11.125) * c5)) / 2 + 1


def in_bounce(step_ratio: float) -> float:
    """
    Ease in using a bounce function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percentage of the current waypoint speed to apply to the character
    """
    return 1 - out_bounce(1 - step_ratio)


def out_bounce(step_ratio: float) -> float:
    """
    Ease out using a bounce function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    n1 = 7.5625
    d1 = 2.75
    if step_ratio < 1 / d1:
        return n1 * step_ratio**2
    elif step_ratio < 2 / d1:
        return n1 * (step_ratio - 1.5 / d1) ** 2 + 0.75
    elif step_ratio < 2.5 / d1:
        return n1 * (step_ratio - 2.25 / d1) ** 2 + 0.9375
    else:
        return n1 * (step_ratio - 2.625 / d1) ** 2 + 0.984375


def in_out_bounce(step_ratio: float) -> float:
    """
    Ease in/out using a bounce function.

    Args:
        step_ratio (float): the ratio of the current step to the maximum steps

    Returns:
        float: 0 <= n <= 1 representing the percent of the inter waypoint distance to travel
    """
    if step_ratio < 0.5:
        return (1 - out_bounce(1 - 2 * step_ratio)) / 2
    else:
        return (1 + out_bounce(2 * step_ratio - 1)) / 2



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rings",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are dispersed and form into spinning rings.",
        description="Characters are dispersed and form into spinning rings.",
        epilog=f"""

Example: terminaltexteffects rings --animation-rate 0.01 --ring-colors ffcc0d ff7326 ff194d bf2669 702a8c --base-color ffffff --ring-gap 6 --spin-duration 200 --disperse-duration 200 --spin-disperse-cycles 3""",
    )
    effect_parser.set_defaults(effect_class=RingsEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--ring-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["ffcc0d", "ff7326", "ff194d", "bf2669", "702a8c"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the rings.",
    )
    effect_parser.add_argument(
        "--base-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when they are not in rings.",
    )
    effect_parser.add_argument(
        "--ring-gap",
        type=argtypes.positive_int,
        default=6,
        help="Distance between rings. Lower distance results in more rings on screen.",
    )
    effect_parser.add_argument(
        "--spin-duration",
        type=argtypes.positive_int,
        default=200,
        help="Number of animation steps for each cycle of the spin phase.",
    )
    effect_parser.add_argument(
        "--disperse-duration",
        type=argtypes.positive_int,
        default=200,
        help="Number of animation steps spent in the dispersed state between spinning cycles.",
    )
    effect_parser.add_argument(
        "--spin-disperse-cycles",
        type=argtypes.positive_int,
        default=3,
        help="Number of times the animation will cycles between spinning rings and dispersed characters.",
    )


class Ring:
    def __init__(
        self,
        radius: int,
        origin: motion.Coord,
        ring_coords: list[motion.Coord],
        ring_gap: int,
        ring_color: str,
        base_color: str,
    ):
        self.radius = radius
        self.origin: motion.Coord = origin
        self.counter_clockwise_coords = ring_coords
        self.clockwise_coords = ring_coords[::-1]
        self.ring_gap = ring_gap
        self.ring_color = ring_color
        self.characters: list[EffectCharacter] = []
        self.character_last_ring_path: dict[EffectCharacter, motion.Path] = {}
        self.rotations = 0
        self.rotation_speed = random.uniform(0.2, 0.5)
        self.gradient = graphics.Gradient(
            [ring_color, "ffffff", ring_color], max(len(self.counter_clockwise_coords) // 3, 1)
        )
        self.gradient_shift = 0
        self.disperse_gradient = graphics.Gradient([self.ring_color, base_color], 7)

    def make_gradient(self, start_color: str, end_color: str, steps: int) -> list[str]:
        gradient = graphics.Gradient([start_color, end_color], steps)
        full_gradient = list(gradient) + list(gradient)[::-1]
        return full_gradient

    def add_character(self, character: EffectCharacter, clockwise: int) -> None:
        # make gradient scene
        gradient_scn = character.animation.new_scene("gradient", is_looping=True)
        for step in self.gradient.spectrum[self.gradient_shift :] + self.gradient.spectrum[: self.gradient_shift]:
            gradient_scn.add_frame(character.input_symbol, 1, color=step)
        self.gradient_shift += 1
        # make rotation waypoints
        ring_paths: list[motion.Path] = []
        character_starting_index = len(self.characters)
        if clockwise:
            coords = self.clockwise_coords
        else:
            coords = self.counter_clockwise_coords
        for coord in coords[character_starting_index:] + coords[:character_starting_index]:
            ring_path = character.motion.new_path(str(len(ring_paths)), speed=self.rotation_speed)
            ring_path.new_waypoint(str(len(ring_path.waypoints)), coord)
            ring_paths.append(ring_path)
        self.character_last_ring_path[character] = ring_paths[0]
        # make disperse scene
        disperse_scn = character.animation.new_scene("disperse", is_looping=False)
        for step in self.disperse_gradient:
            disperse_scn.add_frame(character.input_symbol, 16, color=step)
        character.motion.chain_paths(ring_paths, loop=True)
        self.characters.append(character)

    def make_disperse_waypoints(self, character: EffectCharacter, origin: motion.Coord) -> motion.Path:
        """Make the Path for the character to follow when the ring disperses.

        Args:
            character (EffectCharacter): Character to disperse.
            origin (motion.Coord): Origin coordinate for the character.

        Returns:
            motion.Path: the Path to follow.
        """
        disperse_coords = motion.Motion.find_coords_in_rect(origin, self.ring_gap)
        character.motion.paths.pop("disperse", None)
        disperse_path = character.motion.new_path("disperse", speed=0.1, loop=True)
        for _ in range(5):
            disperse_path.new_waypoint(
                str(len(disperse_path.waypoints)), disperse_coords[random.randrange(0, len(disperse_coords))]
            )
        return disperse_path

    def disperse(self) -> None:
        """Disperse the characters from the ring."""
        for character in self.characters:
            if character.motion.active_path is not None:
                self.character_last_ring_path[character] = character.motion.active_path
            else:
                self.character_last_ring_path[character] = character.motion.paths["0"]
            character.motion.activate_path(self.make_disperse_waypoints(character, character.motion.current_coord))
            character.animation.activate_scene(character.animation.scenes["disperse"])

    def spin(self) -> None:
        for character in self.characters:
            condense_path = character.motion.new_path(str(len(character.motion.paths)), speed=0.1)
            condense_wpt = condense_path.new_waypoint(
                str(len(condense_path.waypoints)),
                self.character_last_ring_path[character].waypoints[0].coord,
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE,
                condense_path,
                EventHandler.Action.ACTIVATE_PATH,
                self.character_last_ring_path[character],
            )
            character.motion.activate_path(condense_path)
            character.animation.activate_scene(character.animation.scenes["gradient"])


class RingsEffect:
    """Effect that forms the characters into rings."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.ring_chars: list[EffectCharacter] = []
        self.non_ring_chars: list[EffectCharacter] = []
        self.rings: dict[int, Ring] = {}
        self.ring_gap = self.args.ring_gap

    def prepare_data(self) -> None:
        """Prepares the data for the effect by building rings and associated animations/waypoints."""
        for character in self.terminal.characters:
            start_scn = character.animation.new_scene("start")
            start_scn.add_frame(character.input_symbol, 1, color=self.args.base_color)
            home_path = character.motion.new_path("home", speed=0.8, ease=easing.out_quad)
            home_wpt = home_path.new_waypoint("home", character.input_coord)
            character.animation.activate_scene(start_scn)
            character.is_active = True
            self.pending_chars.append(character)

        random.shuffle(self.pending_chars)
        # make rings
        for radius in range(1, max(self.terminal.output_area.right, self.terminal.output_area.top), self.ring_gap):
            ring_coords: list[motion.Coord] = []
            circle_coords = motion.Motion.find_coords_on_circle(self.terminal.output_area.center, radius, 7 * radius)
            # check if any part of the ring is in the output area, if not, stop creating rings
            if not any([self.terminal.output_area.coord_in_output_area(coord) for coord in circle_coords]):
                break
            # remove any duplicate coords
            for coord in circle_coords:
                if coord not in ring_coords:
                    ring_coords.append(coord)
            self.rings[radius] = Ring(
                radius,
                self.terminal.output_area.center,
                ring_coords,
                self.ring_gap,
                random.choice(self.args.ring_colors),
                self.args.base_color,
            )
        # assign characters to rings
        ring_count = 0
        for ring in self.rings.values():
            for _ in ring.counter_clockwise_coords:
                if self.pending_chars:
                    next_character = self.pending_chars.pop(0)
                    # set rings to rotate in opposite directions
                    ring.add_character(next_character, clockwise=ring_count % 2)
                    self.ring_chars.append(next_character)
            ring_count += 1

        # make external waypoints for characters not in rings
        for character in self.terminal.characters:
            if character not in self.ring_chars:
                color = random.choice(self.args.ring_colors)
                external_path = character.motion.new_path("external", speed=0.1, ease=easing.out_cubic)
                external_wpt = external_path.new_waypoint("external", self.terminal.random_coord(outside_scope=True))
                # make disperse gradient
                color = random.choice(self.args.ring_colors)
                disperse_gradient = graphics.Gradient([self.args.base_color, color, self.args.base_color], 7)
                disperse_scn = character.animation.new_scene(
                    "disperse", is_looping=False, sync=graphics.SyncMetric.DISTANCE
                )
                for step in disperse_gradient:
                    disperse_scn.add_frame(character.input_symbol, 2, color=step)

                    self.non_ring_chars.append(character)
                    character.event_handler.register_event(
                        EventHandler.Event.PATH_COMPLETE,
                        external_path,
                        EventHandler.Action.SET_CHARACTER_ACTIVATION_STATE,
                        False,
                    )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED,
                    external_path,
                    EventHandler.Action.ACTIVATE_SCENE,
                    disperse_scn,
                )

    def run(self) -> None:
        """Runs the effect."""

        self.prepare_data()
        rings = list(self.rings.values())
        phase = "start"
        initial_disperse_complete = False
        spin_time_remaining = self.args.spin_duration
        disperse_time_remaining = self.args.disperse_duration
        cycles_remaining = self.args.spin_disperse_cycles
        initial_phase_time_remaining = 100
        while phase != "complete":
            if phase == "start":
                if not initial_phase_time_remaining:
                    phase = "disperse"
                else:
                    initial_phase_time_remaining -= 1

            elif phase == "disperse":
                if not initial_disperse_complete:
                    initial_disperse_complete = True
                    for ring in rings:
                        for character in ring.characters:
                            disperse_path = ring.make_disperse_waypoints(
                                character, character.motion.paths["0"].waypoints[0].coord
                            )
                            initial_path = character.motion.new_path("initial", speed=0.3, ease=easing.out_cubic)
                            initial_wpt = initial_path.new_waypoint("initial", disperse_path.waypoints[0].coord)
                            character.event_handler.register_event(
                                EventHandler.Event.PATH_COMPLETE,
                                initial_path,
                                EventHandler.Action.ACTIVATE_PATH,
                                disperse_path,
                            )
                            character.animation.activate_scene(character.animation.scenes["disperse"])
                            character.motion.activate_path(initial_path)
                            self.animating_chars.append(character)

                    for character in self.non_ring_chars:
                        character.motion.activate_path(character.motion.paths["external"])
                        self.animating_chars.append(character)

                else:
                    if not disperse_time_remaining:
                        phase = "spin"
                        cycles_remaining -= 1
                        spin_time_remaining = self.args.spin_duration
                        for ring in rings:
                            ring.spin()
                    else:
                        disperse_time_remaining -= 1

            elif phase == "spin":
                if not spin_time_remaining:
                    if not cycles_remaining:
                        phase = "final"
                        for character in self.terminal.characters:
                            character.is_active = True
                            if "external" in character.motion.paths:
                                self.animating_chars.append(character)
                            character.motion.activate_path(character.motion.paths["home"])
                            character.animation.activate_scene(character.animation.scenes["disperse"])
                    else:
                        disperse_time_remaining = self.args.disperse_duration
                        for ring in rings:
                            ring.disperse()
                        phase = "disperse"

                else:
                    spin_time_remaining -= 1

            elif phase == "final":
                if not self.animating_chars:
                    phase = "complete"

            self.terminal.print()
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



"""This module contains the EffectCharacter class and EventHandler class used to manage the state of a single character from the input data.
"""

import typing
from enum import Enum, auto
from terminaltexteffects.utils import graphics, motion


class EventHandler:
    """Register and handle events related to a character.

    Events related to character state changes (e.g. waypoint reached) can be registered with the EventHandler.
    When an event is triggered, the EventHandler will take the specified action (e.g. activate a waypoint).
    The EventHandler is used by the EffectCharacter class to handle events related to the character.

    Attributes:
        character (EffectCharacter): The character that the EventHandler is handling events for.
        registered_events (dict[tuple[Event, str], list[tuple[Action, str]]]): A dictionary of registered events.
            The key is a tuple of the event and the subject_id (waypoint id/scene id).
            The value is a list of tuples of the action and the action target (waypoint id/scene id).
        layer (int): The layer of the character. The layer determines the order in which characters are printed.
    """

    def __init__(self, character: "EffectCharacter"):
        """Initializes the instance with the EffectCharacter object.

        Args:
            character (EffectCharacter): The character that the EventHandler is handling events for.
        """
        self.character = character
        self.layer: int = 0
        self.registered_events: dict[
            tuple[EventHandler.Event, graphics.Scene | motion.Waypoint | motion.Path],
            list[tuple[EventHandler.Action, graphics.Scene | motion.Waypoint | motion.Path | int | motion.Coord]],
        ] = {}

    class Event(Enum):
        """An Event that can be registered with the EventHandler.

        An Event is triggered when a character reaches a waypoint or an animation scene is activated. Register
        Events with the EventHandler using the register_event method of the EventHandler class.

        Attributes:
            SEGMENT_ENTERED (Event): A path segment has been entered.
            SEGMENT_EXITED (Event): A path segment has been exited.
            PATH_ACTIVATED (Event): A path has been activated.
            PATH_COMPLETE (Event): A path has been completed.
            PATH_HOLDING (Event): A path has entered the holding state.
            SCENE_ACTIVATED (Event): An animation scene has been activated.
            SCENE_COMPLETE (Event): An animation scene has completed.
        """

        SEGMENT_ENTERED = auto()
        SEGMENT_EXITED = auto()
        PATH_ACTIVATED = auto()
        PATH_COMPLETE = auto()
        PATH_HOLDING = auto()
        SCENE_ACTIVATED = auto()
        SCENE_COMPLETE = auto()

    class Action(Enum):
        """Actions that can be taken when an event is triggered.

        An Action is taken when an Event is triggered. Register Actions with the EventHandler using the
        register_event method of the EventHandler class.

        Attributes:
            ACTIVATE_PATH (Action): Activates a path. The action target is the path ID.
            ACTIVATE_SCENE (Action): Activates an animation scene. The action target is the scene ID.
            DEACTIVATE_PATH (Action): Deactivates a path. The action target is the path ID.
            DEACTIVATE_SCENE (Action): Deactivates an animation scene. The action target is the scene ID.
            SET_CHARACTER_ACTIVATION_STATE (Action): Sets the activation state of the character. The action target is the activation state (True/False).
            SET_LAYER (Action): Sets the layer of the character. The action target is the layer number.
            SET_COORDINATE (Action): Sets the coordinate of the character. The action target is the coordinate.
        """

        ACTIVATE_PATH = auto()
        ACTIVATE_SCENE = auto()
        DEACTIVATE_PATH = auto()
        DEACTIVATE_SCENE = auto()
        SET_CHARACTER_ACTIVATION_STATE = auto()
        SET_LAYER = auto()
        SET_COORDINATE = auto()

    def register_event(
        self,
        event: Event,
        caller: graphics.Scene | motion.Waypoint | motion.Path,
        action: Action,
        target: graphics.Scene | motion.Waypoint | motion.Path | int | motion.Coord,
    ) -> None:
        """Registers an event to be handled by the EventHandler.

        Examples:
            >>> character.event_handler.register_event(EventHandler.Event.PATH_COMPLETE, "path_1", EventHandler.Action.ACTIVATE_PATH, "path_2")

        Args:
            event (Event): The event to register.
            subject_id (str): The ID of the event subject (path id/scene id).
            action (Action): The action to take when the event is triggered.
            action_target (str): The ID of the action target.
        """
        new_event = (event, caller)
        new_action = (action, target)
        if new_event not in self.registered_events:
            self.registered_events[new_event] = list()
        self.registered_events[new_event].append(new_action)

    def handle_event(self, event: Event, caller: graphics.Scene | motion.Waypoint | motion.Path) -> None:
        """Handles an event by taking the specified action.

        Examples:
            >>> character.event_handler.handle_event(EventHandler.Event.WAYPOINT_REACHED, "waypoint_1")

        Args:
            event (Event): An event to handle. If the event is not registered, nothing happens.
            caller (graphics.Scene | motion.Waypoint | motion.Path): The object triggering the call.
        """
        action_map = {
            EventHandler.Action.ACTIVATE_PATH: self.character.motion.activate_path,
            EventHandler.Action.ACTIVATE_SCENE: self.character.animation.activate_scene,
            EventHandler.Action.DEACTIVATE_PATH: self.character.motion.deactivate_path,
            EventHandler.Action.DEACTIVATE_SCENE: self.character.animation.deactivate_scene,
            EventHandler.Action.SET_LAYER: lambda layer: setattr(self.character, "layer", layer),
            EventHandler.Action.SET_CHARACTER_ACTIVATION_STATE: lambda state: setattr(
                self.character, "is_active", state
            ),
            EventHandler.Action.SET_COORDINATE: lambda coord: setattr(self.character.motion, "current_coord", coord),
        }

        if (event, caller) not in self.registered_events:
            return
        for event_action in self.registered_events[(event, caller)]:
            action, target = event_action
            action_map[action](target)  # type: ignore


class EffectCharacter:
    """
    A class representing a single character from the input data.

    Attributes:
        input_symbol (str): The symbol for the character in the input data.
        input_coord (motion.Coord): The coordinate of the character in the input data.
        symbol (str): The current symbol for the character, determined by the animation units.
        animation (graphics.Animation): The animation object that controls the character's appearance.
        motion (motion.Motion): The motion object that controls the character's movement.
        event_handler (EventHandler): The event handler object that handles events related to the character.
        is_active (bool): Whether the character is currently active and should be printed to the terminal.
    """

    def __init__(self, symbol: str, input_column: int, input_row: int):
        """Initializes the instance with the input values and the Terminal object.

        Args:
            symbol (str): The symbol for the character in the input data.
            input_column (int): The column of the character in the input data.
            input_row (int): The row of the character in the input data.
        """
        self.input_symbol: str = symbol
        self.input_coord: motion.Coord = motion.Coord(input_column, input_row)
        self.symbol: str = symbol
        self.animation: graphics.Animation = graphics.Animation(self)
        self.motion: motion.Motion = motion.Motion(self)
        self.event_handler: EventHandler = EventHandler(self)
        self.is_active: bool = False
        self.layer: int = 0

    def __hash__(self) -> int:
        return hash(self.input_coord)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, EffectCharacter):
            return NotImplemented
        return self.input_coord == other.input_coord



"""Creates a rain effect where characters fall from the top of the terminal."""

import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.utils import graphics, motion
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "rain",
        formatter_class=argtypes.CustomFormatter,
        help="Rain characters from the top of the output area.",
        description="rain | Rain characters from the top of the output area.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects rain -a 0.01 --rain-colors 39 45 51 21""",
    )
    effect_parser.set_defaults(effect_class=RainEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--rain-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors for the rain drops. Colors are randomly chosen from the list.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.15,
        metavar="(float > 0)",
        help="Falling speed of the rain drops.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_QUART",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class RainEffect:
    """Creates a rain effect where characters fall from the top of the output area."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.group_by_row: dict[int, list[EffectCharacter | None]] = {}

    def prepare_data(self) -> None:
        """Prepares the data for the effect by setting all characters y position to the input height and sorting by target y."""
        if self.args.rain_colors:
            rain_colors = self.args.rain_colors
        else:
            rain_colors = [39, 45, 51, 21, 117, 159]
        rain_characters = ["o", ".", ",", "*", "|"]

        for character in self.terminal.characters:
            raindrop_color = random.choice(rain_colors)
            rain_scn = character.animation.new_scene("rain")
            rain_scn.add_frame(random.choice(rain_characters), 1, color=raindrop_color)
            raindrop_gradient = graphics.Gradient([raindrop_color, self.args.final_color], 7)
            fade_scn = character.animation.new_scene("fade")
            for color in raindrop_gradient:
                fade_scn.add_frame(character.input_symbol, 5, color=color)
            character.animation.activate_scene(rain_scn)
            character.motion.set_coordinate(motion.Coord(character.input_coord.column, self.terminal.output_area.top))
            input_path = character.motion.new_path("input_coord", speed=self.args.movement_speed, ease=self.args.easing)
            input_path.new_waypoint("input_coord", character.input_coord)

            character.event_handler.register_event(
                character.event_handler.Event.PATH_COMPLETE,
                input_path,
                character.event_handler.Action.ACTIVATE_SCENE,
                fade_scn,
            )
            character.motion.activate_path(input_path)
            self.pending_chars.append(character)
        for character in sorted(self.pending_chars, key=lambda c: c.input_coord.row):
            if character.input_coord.row not in self.group_by_row:
                self.group_by_row[character.input_coord.row] = []
            self.group_by_row[character.input_coord.row].append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.pending_chars.clear()
        self.terminal.print()
        while self.group_by_row or self.animating_chars or self.pending_chars:
            if not self.pending_chars and self.group_by_row:
                self.pending_chars.extend(self.group_by_row.pop(min(self.group_by_row.keys())))  # type: ignore
            if self.pending_chars:
                for _ in range(random.randint(1, 3)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop(random.randint(0, len(self.pending_chars) - 1))
                        next_character.is_active = True
                        self.animating_chars.append(next_character)

                    else:
                        break
            self.animate_chars()
            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and getting the next symbol from the animator."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



import argparse
import typing
from terminaltexteffects.utils import easing

EASING_EPILOG = """\
    Easing
    ------
    Note: A prefix must be added to the function name.
    
    All easing functions support the following prefixes:
        IN_  - Ease in
        OUT_ - Ease out
        IN_OUT_ - Ease in and out
        
    Easing Functions
    ----------------
    SINE   - Sine easing
    QUAD   - Quadratic easing
    CUBIC  - Cubic easing
    QUART  - Quartic easing
    QUINT  - Quintic easing
    EXPO   - Exponential easing
    CIRC   - Circular easing
    BACK   - Back easing
    ELASTIC - Elastic easing
    BOUNCE - Bounce easing
    
    Visit: https://easings.net/ for visualizations of the easing functions.
"""


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    """Custom formatter for argparse that combines ArgumentDefaultsHelpFormatter and RawDescriptionHelpFormatter."""

    pass


def valid_color(color_string) -> int | str:
    """Validates that the given argument is a valid color value.

    Args:
        color_string (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Color value is not in range.

    Returns:
        int | str : validated color value
    """
    xterm_min = 0
    xterm_max = 255
    if len(color_string) == 6:
        # Check if the hex value is a valid color
        if not 0 <= int(color_string, 16) <= 16777215:
            raise argparse.ArgumentTypeError(f"invalid color value: {color_string} is not a valid hex color.")
        return color_string
    else:
        # Check if the color is a valid xterm color
        if not xterm_min <= int(color_string) <= xterm_max:
            raise argparse.ArgumentTypeError(f"invalid color value: {color_string} is not a valid xterm color (0-255).")
        return int(color_string)


def valid_duration(duration_arg: str) -> int:
    """Validates that the given argument is a valid duration value.

    Args:
        duration_arg (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Duration value is not in range.

    Returns:
        int: validated duration value
    """
    if int(duration_arg) < 1:
        raise argparse.ArgumentTypeError(
            f"invalid duration value: {duration_arg} is not a valid duration. Duration must be int >= 1."
        )
    return int(duration_arg)


def valid_gap(gap_arg: str) -> int:
    """Validates that the given argument is a valid gap value.

    Args:
        gap_arg (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Gap value is not in range.

    Returns:
        int: validated gap value
    """
    if int(gap_arg) < 0:
        raise argparse.ArgumentTypeError(f"invalid gap value: {gap_arg} is not a valid gap. Gap must be int >= 0.")
    return int(gap_arg)


def valid_animationrate(animationrate_arg: str) -> float:
    """Validates that the given argument is a valid animationrate value.

    Args:
        animationrate_arg (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Animationrate value is not in range.

    Returns:
        float: validated animationrate value
    """
    if float(animationrate_arg) < 0:
        raise argparse.ArgumentTypeError(
            f"invalid animationrate value: {animationrate_arg} is not a valid animationrate. Animationrate must be float >= 0."
        )
    return float(animationrate_arg)


def valid_speed(speed_arg: str) -> float:
    """Validates that the given argument is a valid speed value.

    Args:
        speed_arg (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Speed value is not in range.

    Returns:
        float: validated speed value
    """
    if float(speed_arg) <= 0:
        raise argparse.ArgumentTypeError(
            f"invalid speed value: {speed_arg} is not a valid speed. Speed must be float > 0."
        )
    return float(speed_arg)


def valid_ease(ease_arg: str) -> typing.Callable:
    """Validates that the given argument is a valid easing function.

    Args:
        ease_arg (str): argument to validate

    Raises:
        argparse.ArgumentTypeError: Ease value is not a valid easing function.

    Returns:
        Ease: validated ease value
    """
    easing_func_map = {
        "in_sine": easing.in_sine,
        "out_sine": easing.out_sine,
        "in_out_sine": easing.in_out_sine,
        "in_quad": easing.in_quad,
        "out_quad": easing.out_quad,
        "in_out_quad": easing.in_out_quad,
        "in_cubic": easing.in_cubic,
        "out_cubic": easing.out_cubic,
        "in_out_cubic": easing.in_out_cubic,
        "in_quart": easing.in_quart,
        "out_quart": easing.out_quart,
        "in_out_quart": easing.in_out_quart,
        "in_quint": easing.in_quint,
        "out_quint": easing.out_quint,
        "in_out_quint": easing.in_out_quint,
        "in_expo": easing.in_expo,
        "out_expo": easing.out_expo,
        "in_out_expo": easing.in_out_expo,
        "in_circ": easing.in_circ,
        "out_circ": easing.out_circ,
        "in_out_circ": easing.in_out_circ,
        "in_back": easing.in_back,
        "out_back": easing.out_back,
        "in_out_back": easing.in_out_back,
        "in_elastic": easing.in_elastic,
        "out_elastic": easing.out_elastic,
        "in_out_elastic": easing.in_out_elastic,
        "in_bounce": easing.in_bounce,
        "out_bounce": easing.out_bounce,
        "in_out_bounce": easing.in_out_bounce,
    }

    try:
        return easing_func_map[ease_arg.lower()]
    except KeyError:
        raise argparse.ArgumentTypeError(f"invalid ease value: '{ease_arg}' is not a valid ease.")


def positive_int(arg: str) -> int:
    """Validates that the given argument is a positive integer.

    Args:
        arg (str): argument to validate

    Returns:
        int: validated positive integer
    """
    if int(arg) > 0:
        return int(arg)
    else:
        raise argparse.ArgumentTypeError(f"invalid value: {arg} is not > 0.")


def valid_percent_greater_than_zero(arg: str) -> float:
    """Validates that the given argument is a valid percent value greater than zero.

    Args:
        arg (str): argument to validate

    Returns:
        float: validated percent value
    """
    if 0 < float(arg) <= 100:
        return float(arg)
    else:
        raise argparse.ArgumentTypeError(f"invalid value: {arg} is not > 0 and <= 100.")


def valid_float_between_zero_and_one(arg: str) -> float:
    """Validates that the given argument is a valid float value between zero and one.

    Args:
        arg (str): argument to validate

    Returns:
        float: validated float value
    """
    if 0 <= float(arg) <= 1:
        return float(arg)
    else:
        raise argparse.ArgumentTypeError(f"invalid value: {arg} is not >= 0 and <= 1.")


def ge_zero(arg: str) -> int:
    """Validates that the given argument is an integer greater than or equal to zero.
    Args:
        arg (str): argument to validate

    Returns:
        int: validated integer
    """
    if int(arg) >= 0:
        return int(arg)
    else:
        raise argparse.ArgumentTypeError(f"invalid value: {arg} is not >= 0.")


def is_ascii_or_utf8(s: str) -> bool:
    """Tests if the given string is either ASCII or UTF-8.

    Args:
        s (str): string to test

    Returns:
        bool: True if the string is either ASCII or UTF-8, False otherwise
    """
    try:
        s.encode("ascii")
    except UnicodeEncodeError:
        try:
            s.encode("utf-8")
        except UnicodeEncodeError:
            return False
        else:
            return True
    else:
        return True


def valid_symbol(arg: str) -> str:
    """Validates that the given argument is a valid symbol.

    Args:
        arg (str): argument to validate

    Returns:
        str: validated symbol
    """
    if len(arg) == 1 and is_ascii_or_utf8(arg):
        return arg
    else:
        raise argparse.ArgumentTypeError(
            f"invalid symbol: {arg} is not a valid symbol. Must be a single ASCII/UTF-8 character."
        )



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.utils import graphics, motion
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "bouncyballs",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are bouncy balls falling from the top of the output area.",
        description="bouncyball | Characters are bouncy balls falling from the top of the output area.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects bouncyball -a 0.01 --ball-colors 00ff00 ff0000 0000ff --final-color ffffff --ball-delay 15 --movement-speed 0.25 --easing OUT_BOUNCE""",
    )
    effect_parser.set_defaults(effect_class=BouncyBallsEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time between animation steps. ",
    )
    effect_parser.add_argument(
        "--ball-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated list of colors from which ball colors will be randomly selected. If no colors are provided, the colors are random.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--ball-delay",
        type=argtypes.positive_int,
        default=15,
        metavar="(int > 0)",
        help="Number of animation steps between ball drops, increase to reduce ball drop rate.",
    ),
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.25,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="OUT_BOUNCE",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class BouncyBallsEffect:
    """Effect that displays the text as bouncy balls falling from the top of the output area."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.group_by_row: dict[int, list[EffectCharacter | None]] = {}

    def prepare_data(self) -> None:
        """Prepares the data for the effect by assigning colors and waypoints and
        organizing the characters by row."""
        ball_symbols = ("*", "o", "O", "0", ".")
        for character in self.terminal.characters:
            character.is_active = False
            if self.args.ball_colors:
                color = random.choice(self.args.ball_colors)
            else:
                color = graphics.Animation.random_color()
            symbol = random.choice(ball_symbols)
            ball_scene = character.animation.new_scene("ball")
            ball_scene.add_frame(symbol, 1, color=color)
            final_scene = character.animation.new_scene("final")
            for step in graphics.Gradient([color, self.args.final_color], 12):
                final_scene.add_frame(
                    character.input_symbol,
                    10,
                    color=step,
                )
            character.motion.set_coordinate(motion.Coord(character.input_coord.column, self.terminal.output_area.top))
            input_coord_path = character.motion.new_path(
                "input_pth", speed=self.args.movement_speed, ease=self.args.easing
            )
            input_coord_waypoint = input_coord_path.new_waypoint("input_coord_wpt", character.input_coord)
            character.motion.activate_path(input_coord_path)
            character.animation.activate_scene(ball_scene)
            character.event_handler.register_event(
                character.event_handler.Event.PATH_COMPLETE,
                input_coord_path,
                character.event_handler.Action.ACTIVATE_SCENE,
                final_scene,
            )
            self.pending_chars.append(character)
        for character in sorted(self.pending_chars, key=lambda c: c.input_coord.row):
            if character.input_coord.row not in self.group_by_row:
                self.group_by_row[character.input_coord.row] = []
            self.group_by_row[character.input_coord.row].append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.pending_chars.clear()
        ball_delay = 0
        while self.group_by_row or self.animating_chars or self.pending_chars:
            if not self.pending_chars and self.group_by_row:
                self.pending_chars.extend(self.group_by_row.pop(min(self.group_by_row.keys())))  # type: ignore
            if self.pending_chars and ball_delay == 0:
                for _ in range(random.randint(1, 5)):
                    if self.pending_chars:
                        next_character = self.pending_chars.pop(random.randint(0, len(self.pending_chars) - 1))
                        next_character.is_active = True
                        self.animating_chars.append(next_character)
                    else:
                        break
                ball_delay = self.args.ball_delay
            ball_delay -= 1
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        """Animates the characters by calling the tween method and printing the characters to the terminal."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "expand",
        formatter_class=argtypes.CustomFormatter,
        help="Expands the text from a single point.",
        description="expand | Expands the text from a single point.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects expand -a 0.01""",
    )
    effect_parser.set_defaults(effect_class=ExpandEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_QUART",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class ExpandEffect:
    """Effect that draws the characters expanding from a single point."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by starting all of the characters from a point in the middle of the input data."""

        for character in self.terminal.characters:
            character.motion.set_coordinate(self.terminal.output_area.center)
            input_coord_path = character.motion.new_path(
                "input_coord",
                speed=self.args.movement_speed,
                ease=self.args.easing,
            )
            input_coord_wpt = input_coord_path.new_waypoint("input_coord", character.input_coord)
            character.is_active = True
            character.motion.activate_path(input_coord_path)
            self.animating_chars.append(character)
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED, input_coord_path, EventHandler.Action.SET_LAYER, 1
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, input_coord_path, EventHandler.Action.SET_LAYER, 0
            )

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.terminal.print()
        while self.animating_chars:
            self.animate_chars()
            self.animating_chars = [
                animating for animating in self.animating_chars if not animating.motion.movement_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        for animating_char in self.animating_chars:
            animating_char.motion.move()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes, motion


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "middleout",
        formatter_class=argtypes.CustomFormatter,
        help="Text expands in a single row or column in the middle of the output area then out.",
        description="Text expands in a single row or column in the middle of the output area then out.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects middleout -a 0.01 --expand-direction horizontal --center-movement-speed 0.35 
         --full-movement-speed 0.35 --center-easing IN_OUT_SINE --full-easing IN_OUT_SINE 
         --center-expand-color 00ff00 --full-expand-color 0000ff --starting-color ff0000""",
    )
    effect_parser.set_defaults(effect_class=MiddleoutEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--starting-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the initial text in the center of the output area.",
    )
    effect_parser.add_argument(
        "--center-expand-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the row or column as it expands.",
    )
    effect_parser.add_argument(
        "--full-expand-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the full text as it expands.",
    )
    effect_parser.add_argument(
        "--expand-direction",
        default="vertical",
        choices=["vertical", "horizontal"],
        help="Direction the text will expand.",
    )
    effect_parser.add_argument(
        "--center-movement-speed",
        type=argtypes.valid_speed,
        default=0.35,
        metavar="(float > 0)",
        help="Speed of the characters during the initial expansion of the center vertical/horiztonal line. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--full-movement-speed",
        type=argtypes.valid_speed,
        default=0.35,
        metavar="(float > 0)",
        help="Speed of the characters during the final full expansion. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--center-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for initial expansion.",
    )
    effect_parser.add_argument(
        "--full-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for full expansion.",
    )


class MiddleoutEffect:
    """Effect that expands a single row and column followed by the rest of the output area."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.center_gradient = graphics.Gradient([self.args.starting_color, self.args.center_expand_color], 10)
        self.full_gradient = graphics.Gradient([self.args.center_expand_color, self.args.full_expand_color], 10)

    def prepare_data(self) -> None:
        """Prepares the data for the effect."""
        for character in self.terminal.characters:
            character.motion.set_coordinate(self.terminal.output_area.center)
            # setup waypoints
            if self.args.expand_direction == "vertical":
                column = character.input_coord.column
                row = self.terminal.output_area.center_row
            else:
                column = self.terminal.output_area.center_column
                row = character.input_coord.row
            center_path = character.motion.new_path(
                "center", speed=self.args.center_movement_speed, ease=self.args.center_easing
            )
            center_waypoint = center_path.new_waypoint("center", motion.Coord(column, row))
            full_path = character.motion.new_path(
                "full", speed=self.args.full_movement_speed, ease=self.args.full_easing
            )
            full_waypoint = full_path.new_waypoint("full", character.input_coord)

            # setup scenes
            center_scene = character.animation.new_scene("center")
            full_scene = character.animation.new_scene("full")
            for step in self.center_gradient:
                center_scene.add_frame(character.input_symbol, 2, color=step)
            for step in self.full_gradient:
                full_scene.add_frame(character.input_symbol, 2, color=step)
            character.animation.activate_scene(center_scene)

            # initialize character state
            character.motion.activate_path(center_path)
            character.animation.activate_scene(center_scene)
            character.is_active = True
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        final = False
        while self.pending_chars or self.animating_chars:
            if all([character.motion.active_path is None for character in self.animating_chars]):
                final = True
                for character in self.animating_chars:
                    character.motion.activate_path(character.motion.paths["full"])
                    character.animation.activate_scene(character.animation.scenes["full"])
            self.terminal.print()
            self.animate_chars()
            if final:
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.animation.active_scene_is_complete()
                    or not animating_char.motion.movement_is_complete()
                ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random

from terminaltexteffects.utils import motion
import terminaltexteffects.utils.argtypes as argtypes
import terminaltexteffects.utils.terminal as terminal
from terminaltexteffects.base_character import EffectCharacter, EventHandler


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "scattered",
        formatter_class=argtypes.CustomFormatter,
        help="Move the characters into place from random starting locations.",
        description="scattered | Move the characters into place from random starting locations.",
        epilog=f"""{argtypes.EASING_EPILOG}
        
Example: terminaltexteffects scattered -a 0.01""",
    )
    effect_parser.set_defaults(effect_class=ScatteredEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Time between animation steps.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=0.5,
        metavar="(float > 0)",
        help="Movement speed of the characters. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_BACK",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class ScatteredEffect:
    """Effect that moves the characters into position from random starting locations."""

    def __init__(self, terminal: terminal.Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by scattering the characters within range of the input width and height."""
        for character in self.terminal.characters:
            if self.terminal.output_area.right < 2 or self.terminal.output_area.top < 2:
                character.motion.set_coordinate(motion.Coord(1, 1))
            else:
                character.motion.set_coordinate(self.terminal.random_coord())
            input_coord_path = character.motion.new_path(
                "input_pth", speed=self.args.movement_speed, ease=self.args.easing
            )
            input_coord_path.new_waypoint("input_coord_wpt", character.input_coord)
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED, input_coord_path, EventHandler.Action.SET_LAYER, 1
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, input_coord_path, EventHandler.Action.SET_LAYER, 0
            )
            character.motion.activate_path(input_coord_path)
            character.is_active = True
            self.animating_chars.append(character)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        self.terminal.print()
        while self.pending_chars or self.animating_chars:
            self.animate_chars()
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.motion.movement_is_complete()
            ]
            self.terminal.print()

    def animate_chars(self) -> None:
        for animating_char in self.animating_chars:
            animating_char.motion.move()



"""This file contains ANSI escape codes for terminal formatting."""


def DEC_SAVE_CURSOR_POSITION() -> str:
    """Saves the cursor position using DEC sequence.

    Returns:
        str: ANSI escape code
    """
    return "\0337"


def DEC_RESTORE_CURSOR_POSITION() -> str:
    """Restores the cursor position using DEC sequence.

    Returns:
        str: ANSI escape code
    """
    return "\0338"


def HIDE_CURSOR() -> str:
    """Hides the cursor.

    Returns:
        str: ANSI escape code
    """
    return "\033[?25l"


def SHOW_CURSOR() -> str:
    """Shows the cursor.

    Returns:
        str: ANSI escape code
    """
    return "\033[?25h"


def MOVE_CURSOR_UP(y: int) -> str:
    """Moves the cursor up y lines.

    Args:
        y (int): number of lines to move up

    Returns:
        str: ANSI escape code
    """
    return f"\033[{y}A"


def MOVE_CURSOR_TO_COLUMN(x: int) -> str:
    """Moves the cursor to the x column.

    Args:
        x (int): column number

    Returns:
        str: ANSI escape code
    """
    return f"\033[{x}G"


def RESET_ALL() -> str:
    """Resets all formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[0m"


def APPLY_BOLD() -> str:
    """Applies bold formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[1m"


def APPLY_DIM() -> str:
    """Applies dim formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[2m"


def APPLY_ITALIC() -> str:
    """Applies italic formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[3m"


def APPLY_UNDERLINE() -> str:
    """Applies underline formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[4m"


def APPLY_BLINK() -> str:
    """Applies blink formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[5m"


def APPLY_REVERSE() -> str:
    """Applies reverse formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[7m"


def APPLY_HIDDEN() -> str:
    """Applies hidden formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[8m"


def APPLY_STRIKETHROUGH() -> str:
    """Applies strikethrough formatting.

    Returns:
        str: ANSI escape code
    """
    return "\033[9m"



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes, motion


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "unstable",
        formatter_class=argtypes.CustomFormatter,
        help="Spawn characters jumbled, explode them to the edge of the output area, then reassemble them in the correct layout.",
        description="unstable | Spawn characters jumbled, explode them to the edge of the output area, then reassemble them in the correct layout.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects unstable -a 0.01 --initial-color ffffff --unstable-color ff9200 --final-color ffffff --explosion-ease OUT_EXPO --explosion-speed 0.75 --reassembly-ease OUT_EXPO --reassembly-speed 0.75""",
    )
    effect_parser.set_defaults(effect_class=UnstableEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--initial-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when the effect starts.",
    )
    effect_parser.add_argument(
        "--unstable-color",
        type=argtypes.valid_color,
        default="ff9200",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color transitioned to as the characters become unstable.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color transitioned to as the characters reassemble.",
    )
    effect_parser.add_argument(
        "--explosion-ease",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement during the explosion.",
    )
    effect_parser.add_argument(
        "--explosion-speed",
        type=argtypes.valid_speed,
        default=0.75,
        metavar="(float > 0)",
        help="Speed of characters during explosion. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--reassembly-ease",
        default="OUT_EXPO",
        type=argtypes.valid_ease,
        help="Easing function to use for character reassembly.",
    )
    effect_parser.add_argument(
        "--reassembly-speed",
        type=argtypes.valid_speed,
        default=0.75,
        metavar="(float > 0)",
        help="Speed of characters during reassembly. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )


class UnstableEffect:
    """Effect that spawns characters jumbled, explodes them to the edge of the output area,
    then reassembles them in the correct layout."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.jumbled_coords: dict[EffectCharacter, motion.Coord] = dict()

    def prepare_data(self) -> None:
        """Prepares the data for the effect by jumbling the character positions and
        choosing a location on the perimeter of the output area for the character to travel
        after exploding. Creates all waypoints and scenes for the characters."""
        character_coords = [character.input_coord for character in self.terminal.characters]
        for character in self.terminal.characters:
            pos = random.randint(0, 3)
            match pos:
                case 0:
                    col = self.terminal.output_area.left
                    row = random.randint(1, self.terminal.output_area.top)
                case 1:
                    col = self.terminal.output_area.right
                    row = random.randint(1, self.terminal.output_area.top)
                case 2:
                    col = random.randint(1, self.terminal.output_area.right)
                    row = self.terminal.output_area.bottom
                case 3:
                    col = random.randint(1, self.terminal.output_area.right)
                    row = self.terminal.output_area.top
            jumbled_coord = character_coords.pop(random.randint(0, len(character_coords) - 1))
            self.jumbled_coords[character] = jumbled_coord
            character.motion.set_coordinate(jumbled_coord)
            explosion_path = character.motion.new_path("explosion", speed=0.75, ease=self.args.explosion_ease)
            explosion_wpt = explosion_path.new_waypoint("explosion", motion.Coord(col, row))
            reassembly_path = character.motion.new_path("reassembly", speed=0.75, ease=self.args.reassembly_ease)
            reassembly_wpt = reassembly_path.new_waypoint("reassembly", character.input_coord)
            unstable_gradient = graphics.Gradient([self.args.initial_color, self.args.unstable_color], 25)
            rumble_scn = character.animation.new_scene("rumble")
            for step in unstable_gradient:
                rumble_scn.add_frame(character.input_symbol, 10, color=step)
            final_color = graphics.Gradient([self.args.unstable_color, self.args.final_color], 12)
            final_scn = character.animation.new_scene("final")
            for step in final_color:
                final_scn.add_frame(character.input_symbol, 5, color=step)
            character.animation.activate_scene(rumble_scn)
            character.is_active = True

    def move_all_to_waypoint(self, path_id) -> None:
        for character in self.terminal.characters:
            if path_id == "reassembly":
                character.animation.activate_scene(character.animation.scenes["final"])
            self.animating_chars.append(character)
            character.motion.activate_path(character.motion.paths[path_id])
        while self.animating_chars:
            self.terminal.print()
            self.animate_chars()
            if path_id == "reassembly":
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.motion.current_coord
                    == animating_char.motion.paths[path_id].waypoints[0].coord
                    or not animating_char.animation.active_scene_is_complete()
                ]
            else:
                self.animating_chars = [
                    animating_char
                    for animating_char in self.animating_chars
                    if not animating_char.motion.current_coord
                    == animating_char.motion.paths[path_id].waypoints[0].coord
                ]

    def rumble(self) -> None:
        max_rumble_steps = 250
        current_rumble_steps = 0
        rumble_mod_delay = 20
        while current_rumble_steps < max_rumble_steps:
            if current_rumble_steps > 30 and current_rumble_steps % rumble_mod_delay == 0:
                row_offset = random.choice([-1, 0, 1])
                column_offset = random.choice([-1, 0, 1])
                for character in self.terminal.characters:
                    character.motion.set_coordinate(
                        motion.Coord(
                            character.motion.current_coord.column + column_offset,
                            character.motion.current_coord.row + row_offset,
                        )
                    )
                    character.animation.step_animation()
                self.terminal.print()
                for character in self.terminal.characters:
                    character.motion.set_coordinate(self.jumbled_coords[character])
                rumble_mod_delay -= 1
                rumble_mod_delay = max(rumble_mod_delay, 1)
            else:
                for character in self.terminal.characters:
                    character.animation.step_animation()
                self.terminal.print()

            current_rumble_steps += 1

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        explosion_hold_time = 50
        self.rumble()
        self.move_all_to_waypoint("explosion")
        while explosion_hold_time:
            self.terminal.print()
            self.animate_chars()
            explosion_hold_time -= 1
        self.move_all_to_waypoint("reassembly")

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "swarm",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are grouped into swarms and move around the terminal before settling into position.",
        description="Characters are grouped into swarms and move around the terminal before settling into position.",
        epilog=f"""Example: terminaltexteffects swarm -a 0.01 --base-color 0f74fa --flash-color f2ea79 --final-color 96ed89 --swarm-size 5 --swarm-coordination 80""",
    )
    effect_parser.set_defaults(effect_class=SwarmEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--base-color",
        type=argtypes.valid_color,
        default="0f74fa",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when not flashing.",
    )
    effect_parser.add_argument(
        "--flash-color",
        type=argtypes.valid_color,
        default="f2ea79",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the character flash. Characters flash when moving.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="96ed89",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the characters when they reach their input coordinate.",
    )
    effect_parser.add_argument(
        "--swarm-size",
        type=argtypes.valid_percent_greater_than_zero,
        default=5,
        help="Percent of total characters in each swarm. Must be between 1 and 100.",
    )
    effect_parser.add_argument(
        "--swarm-coordination",
        type=argtypes.valid_percent_greater_than_zero,
        default=80,
        help="Percent of characters in a swarm that move as a group. Must be between 1 and 100.",
    )


class SwarmEffect:
    """Characters behave with swarm characteristics before flying into position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.swarms: list[list[EffectCharacter]] = []
        self.swarm_size: int = max(round(len(self.terminal.characters) * (self.args.swarm_size / 100)), 1)

    def make_swarms(self) -> None:
        unswarmed_characters = list(self.terminal.characters[::-1])
        while unswarmed_characters:
            new_swarm: list[EffectCharacter] = []
            for _ in range(self.swarm_size):
                if unswarmed_characters:
                    new_swarm.append(unswarmed_characters.pop())
                else:
                    break
            self.swarms.append(new_swarm)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by creating swarms of characters and setting waypoints and animations."""
        self.make_swarms()
        final_gradient = graphics.Gradient([self.args.base_color, self.args.final_color], 10)
        swarm_gradient = graphics.Gradient([self.args.base_color, self.args.flash_color], 7)
        flash_list = [self.args.flash_color for _ in range(10)]
        swarm_gradient_mirror = list(swarm_gradient) + flash_list + list(swarm_gradient)[::-1]
        for swarm in self.swarms:
            swarm_area_coordinate_map: dict[motion.Coord, list[motion.Coord]] = {}
            swarm_spawn = self.terminal.random_coord(outside_scope=True)
            swarm_areas: list[motion.Coord] = []
            swarm_area_count = random.randint(2, 4)
            # create areas where characters will swarm
            while len(swarm_areas) < swarm_area_count:
                # get random coord within inner 90% of terminal
                col = random.randint(round(self.terminal.input_width * 0.1), round(self.terminal.input_width * 0.9) + 1)
                row = random.randint(
                    round(self.terminal.output_area.top * 0.1), round(self.terminal.output_area.top * 0.9) + 1
                )
                area = motion.Coord(col, row)
                if area not in swarm_areas:
                    swarm_areas.append(area)
                    # swarm area radius is a function of the terminal size, approximately 1/4 of the smallest dimension
                    swarm_area_coordinate_map[area] = motion.Motion.find_coords_in_circle(
                        area, max(min(self.terminal.output_area.right, self.terminal.output_area.top) // 6, 1)
                    )

            # assign characters waypoints for swarm areas and inner waypoints within the swarm areas
            for character in swarm:
                swarm_area_count = 0
                character.motion.set_coordinate(swarm_spawn)
                flash_scn = character.animation.new_scene("speed", sync=graphics.SyncMetric.DISTANCE)
                for step in swarm_gradient_mirror:
                    flash_scn.add_frame(character.input_symbol, 1, color=step)
                for _, swarm_area_coords in swarm_area_coordinate_map.items():
                    swarm_area_name = f"{swarm_area_count}_swarm_area"
                    swarm_area_count += 1
                    origin_path = character.motion.new_path(swarm_area_name, speed=0.25, ease=easing.out_sine)
                    origin_wpt = origin_path.new_waypoint(swarm_area_name, random.choice(swarm_area_coords))
                    character.event_handler.register_event(
                        EventHandler.Event.PATH_ACTIVATED, origin_path, EventHandler.Action.ACTIVATE_SCENE, flash_scn
                    )
                    character.event_handler.register_event(
                        EventHandler.Event.PATH_ACTIVATED, origin_path, EventHandler.Action.SET_LAYER, 1
                    )
                    character.event_handler.register_event(
                        EventHandler.Event.PATH_COMPLETE,
                        origin_path,
                        EventHandler.Action.DEACTIVATE_SCENE,
                        flash_scn,
                    )
                    inner_paths = 0
                    total_inner_paths = 2
                    while inner_paths < total_inner_paths:
                        next_coord = random.choice(swarm_area_coords)
                        inner_paths += 1
                        inner_path = character.motion.new_path(
                            str(len(character.motion.paths)), speed=0.1, ease=easing.in_out_sine
                        )
                        inner_path.new_waypoint(str(len(character.motion.paths)), next_coord)
                # create landing waypoint and scene
                input_path = character.motion.new_path("input_coord", speed=0.3, ease=easing.in_out_quad)
                input_wpt = input_path.new_waypoint("input_coord", character.input_coord)
                input_scn = character.animation.new_scene("input")
                for step in final_gradient:
                    input_scn.add_frame(character.input_symbol, 3, color=step)
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE, input_path, EventHandler.Action.ACTIVATE_SCENE, input_scn
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE, input_path, EventHandler.Action.SET_LAYER, 0
                )
                # create flash effect when moving between waypoints
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED, input_path, EventHandler.Action.ACTIVATE_SCENE, flash_scn
                )
                # chain waypoints
                character.motion.chain_paths(list(character.motion.paths.values()))
                # last_wpt = None
                # for wpt in character.motion.waypoints.values():
                #     if last_wpt:
                #         character.event_handler.register_event(
                #             EventHandler.Event.WAYPOINT_COMPLETE, last_wpt, EventHandler.Action.ACTIVATE_WAYPOINT, wpt
                #         )
                #     last_wpt = wpt

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        call_next = True
        active_swarm_area = "0_swarm_area"
        while self.swarms or self.animating_chars:
            if self.swarms and call_next:
                call_next = False
                current_swarm = self.swarms.pop()
                active_swarm_area = "0_swarm_area"
                for character in current_swarm:
                    character.motion.activate_path(character.motion.paths["0_swarm_area"])
                    character.is_active = True
                    self.animating_chars.append(character)
            self.terminal.print()
            self.animate_chars()
            if len(self.animating_chars) < len(current_swarm):
                call_next = True
            if current_swarm:
                for character in current_swarm:
                    if (
                        character.motion.active_path
                        and character.motion.active_path.path_id != active_swarm_area
                        and "swarm_area" in character.motion.active_path.path_id
                        and int(character.motion.active_path.path_id[0]) > int(active_swarm_area[0])
                    ):
                        active_swarm_area = character.motion.active_path.path_id
                        for other in current_swarm:
                            if other is not character and random.random() < (self.args.swarm_coordination / 100):
                                other.motion.activate_path(other.motion.paths[active_swarm_area])
                        break

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "effect_name",
        formatter_class=argtypes.CustomFormatter,
        help="effect_description",
        description="effect_description",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: effect_example""",
    )
    effect_parser.set_defaults(effect_class=NamedEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--color-single",
        type=argtypes.valid_color,
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the ___.",
    )
    effect_parser.add_argument(
        "--color-list",
        type=argtypes.valid_color,
        nargs="*",
        default=0,
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Space separated, unquoted, list of colors for the ___.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--movement-speed",
        type=argtypes.valid_speed,
        default=1,
        metavar="(float > 0)",
        help="Speed of the ___. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement.",
    )


class NamedEffect:
    """Effect that ___."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""

        for character in self.terminal.characters:
            pass
            # do something with the data if needed (sort, adjust positions, etc)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_chars or self.animating_chars:
            self.terminal.print()
            self.animate_chars()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



"""Converts xterm color codes and hex colors into ANSI escape sequences."""


RESET = "\x1b[0m"


def _hex_to_int(hex_color: str) -> list[int]:
    """
    Convert a hex color string into a list of ints.

    Parameters
    ----------
    hex_color : str : Hex color string. '#' is optional.

    Returns
    -------
    List[int] : [RED, GREEN, BLUE] list of color integers
    """
    hex_color = hex_color.strip("#")
    ints = [int(hex_color[i : i + 2], 16) for i in range(0, 6, 2)]
    return ints


def _color(color_code: str | int, location: int) -> str:
    """
    Returns an ANSI escape sequence to color the foreground/background of text following the returned sequence.
    Parameters
    ----------
    color_code : Union[str, int]
    location : int

    Returns
    -------
    str
    """

    if isinstance(color_code, str):
        color_ints = _hex_to_int(color_code)
        sequence = f"\x1b[{location};2;{color_ints[0]};{color_ints[1]};{color_ints[2]}m"
    elif isinstance(color_code, int):
        if color_code not in range(0, 256):
            raise ValueError("Got color code ({color_code}): xterm color codes must be an integer: 0 <= n <= 255")
        else:
            sequence = f"\x1b[{location};5;{color_code}m"
    else:
        raise ValueError(
            f"Got color code ({color_code}): Color must be either hex string #000000 -> #FFFFFF or"
            f" int xterm color code 0 <= n <= 255"
        )
    return sequence


def fg(color_code: str | int) -> str:
    """Returns an ANSI escape sequence to color the foreground of text following the returned sequence.

    Parameters
    ----------
    color_code : Union[str, int] : Hex color string or xterm color int 0 <= n <= 255

    Returns
    -------
    str : ANSI escape sequence
    """
    sequence = _color(color_code, 38)
    return sequence


def bg(color_code: str | int) -> str:
    """Returns an ANSI escape sequence to color the background of text following the returned sequence.

    Parameters
    ----------
    color_code : Union[str, int] : Hex color string or xterm color int 0 <= n <= 255

    Returns
    -------
    str : ANSI escape sequence
    """
    sequence = _color(color_code, 48)
    return sequence



import argparse
import random
import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, motion, argtypes, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "blackhole",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are consumed by a black hole and explode outwards.",
        description="Characters are consumed by a black hole and explode outwards.",
        epilog=f"""
Example: terminaltexteffects blackhole -a 0.01 --star-colors ffcc0d ff7326 ff194d bf2669 702a8c 049dbf --final-color 00a7c2""",
    )
    effect_parser.set_defaults(effect_class=BlackholeEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--blackhole-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the stars that comprise the blackhole border.",
    )
    effect_parser.add_argument(
        "--star-colors",
        type=argtypes.valid_color,
        nargs="*",
        default=["ffcc0d", "ff7326", "ff194d", "bf2669", "702a8c", "049dbf"],
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="List of colors from which character colors will be chosen and applied after the explosion, but before the cooldown to final color.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="00a7c2",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Final color that characters will shift to after exploding.",
    )


class BlackholeEffect:
    """Effect that creates a blackhole in a starfield, consumes the stars, and explodes the input characters out to position."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.blackhole_chars: list[EffectCharacter] = []
        self.awaiting_consumption_chars: list[EffectCharacter] = []
        self.blackhole_radius = max(
            min(round(self.terminal.output_area.right * 0.3), round(self.terminal.output_area.top * 0.3)), 3
        )

    def prepare_blackhole(self) -> None:
        star_symbols = [
            "*",
            "✸",
            "✺",
            "✹",
            "✷",
            "✵",
            "✴",
            "✶",
            "⋆",
            "'",
            ".",
            "⬫",
            "⬪",
            "⬩",
            "⬨",
            "⬧",
            "⬦",
            "⬥",
        ]
        starfield_colors = graphics.Gradient(["4a4a4d", "ffffff"], 6).spectrum
        gradient_map = {}
        for color in starfield_colors:
            gradient_map[color] = graphics.Gradient([color, "000000"], 10)
        available_chars = list(self.terminal.characters)
        while len(self.blackhole_chars) < self.blackhole_radius * 3 and available_chars:
            self.blackhole_chars.append(available_chars.pop(random.randrange(0, len(available_chars))))
        black_hole_ring_positions = motion.Motion.find_coords_on_circle(
            self.terminal.output_area.center, self.blackhole_radius, len(self.blackhole_chars)
        )
        for position_index, character in enumerate(self.blackhole_chars):
            starting_pos = black_hole_ring_positions[position_index]
            blackhole_path = character.motion.new_path("blackhole", speed=0.5, ease=easing.in_out_sine)
            blackhole_wpt = blackhole_path.new_waypoint("blackhole", starting_pos)
            blackhole_scn = character.animation.new_scene("blackhole")
            blackhole_scn.add_frame("✸", 1, color=self.args.blackhole_color)
            character.event_handler.register_event(
                EventHandler.Event.PATH_ACTIVATED, blackhole_path, EventHandler.Action.SET_LAYER, 1
            )
            # make rotation waypoints
            blackhole_rotation_path = character.motion.new_path("blackhole_rotation", speed=0.2, loop=True)
            for coord in black_hole_ring_positions[position_index:] + black_hole_ring_positions[:position_index]:
                blackhole_rotation_path.new_waypoint(str(len(blackhole_rotation_path.waypoints)), coord)
        for character in self.terminal.characters:
            character.is_active = True
            starting_scn = character.animation.new_scene("starting")
            star_symbol = random.choice(star_symbols)
            star_color = random.choice(starfield_colors)
            starting_scn.add_frame(star_symbol, 1, color=star_color)
            character.animation.activate_scene(starting_scn)
            if character not in self.blackhole_chars:
                starfield_coord = self.terminal.random_coord()
                character.motion.set_coordinate(starfield_coord)
                if starfield_coord.row > self.terminal.output_area.center_row:
                    if starfield_coord.column in range(
                        round(self.terminal.output_area.right * 0.4), round(self.terminal.output_area.right * 0.7)
                    ):
                        # if within the top center 40% of the screen
                        control_point = motion.Coord(self.terminal.output_area.center.column, starfield_coord.row)
                    else:
                        control_point = motion.Coord(starfield_coord.column, self.terminal.output_area.center_row)

                elif starfield_coord.row < self.terminal.output_area.center_row:
                    if starfield_coord.column in range(
                        round(self.terminal.output_area.right * 0.4), round(self.terminal.output_area.right * 0.7)
                    ):
                        # if within the bottom center 40% of the screen
                        control_point = motion.Coord(self.terminal.output_area.center.column, starfield_coord.row)
                    else:
                        control_point = motion.Coord(starfield_coord.column, self.terminal.output_area.center_row)
                else:
                    control_point = self.terminal.output_area.center
                singularity_path = character.motion.new_path("singularity", speed=0.3, ease=easing.in_expo)
                singularity_wpt = singularity_path.new_waypoint(
                    "singularity",
                    self.terminal.output_area.center,
                    bezier_control=control_point,
                )
                consumed_scn = character.animation.new_scene("consumed")
                for color in gradient_map[star_color]:
                    consumed_scn.add_frame(star_symbol, 1, color=color)
                consumed_scn.add_frame(" ", 1)
                consumed_scn.sync = graphics.SyncMetric.DISTANCE
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED, singularity_path, EventHandler.Action.SET_LAYER, 2
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_ACTIVATED,
                    singularity_path,
                    EventHandler.Action.ACTIVATE_SCENE,
                    consumed_scn,
                )
                self.awaiting_consumption_chars.append(character)
        random.shuffle(self.awaiting_consumption_chars)

    def rotate_blackhole(self) -> None:
        for character in self.blackhole_chars:
            character.motion.activate_path(character.motion.paths["blackhole_rotation"])
            self.animating_chars.append(character)

    def collapse_blackhole(self) -> None:
        black_hole_ring_positions = motion.Motion.find_coords_on_circle(
            self.terminal.output_area.center, self.blackhole_radius + 3, len(self.blackhole_chars)
        )
        unstable_symbols = ["◦", "◎", "◉", "●", "◉", "◎", "◦"]
        point_char_made = False
        for character in self.blackhole_chars:
            next_pos = black_hole_ring_positions.pop(0)
            expand_path = character.motion.new_path("expand", speed=0.1, ease=easing.in_expo)
            expand_wpt = expand_path.new_waypoint("expand", next_pos)
            collapse_path = character.motion.new_path("collapse", speed=0.3, ease=easing.in_expo)
            collapse_wpt = collapse_path.new_waypoint("collapse", self.terminal.output_area.center)
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, expand_path, EventHandler.Action.ACTIVATE_PATH, collapse_path
            )
            if not point_char_made:
                point_scn = character.animation.new_scene("point")
                for _ in range(3):
                    for symbol in unstable_symbols:
                        point_scn.add_frame(symbol, 6, color=random.choice(self.args.star_colors))
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE, collapse_path, EventHandler.Action.ACTIVATE_SCENE, point_scn
                )
                character.event_handler.register_event(
                    EventHandler.Event.PATH_COMPLETE, collapse_path, EventHandler.Action.SET_LAYER, 3
                )
                point_char_made = True

            character.motion.activate_path(expand_path)
            self.animating_chars.append(character)

    def explode_singularity(self) -> None:
        star_colors = ["ffcc0d", "ff7326", "ff194d", "bf2669", "702a8c" "049dbf"]
        for character in self.terminal.characters:
            nearby_coord = motion.Motion.find_coords_on_circle(character.input_coord, 3, 5)[random.randrange(0, 5)]
            nearby_path = character.motion.new_path("nearby", speed=random.randint(2, 3) / 10, ease=easing.out_expo)
            nearby_wpt = nearby_path.new_waypoint("nearby_wpt", nearby_coord)
            input_path = character.motion.new_path("input", speed=random.randint(3, 5) / 100, ease=easing.in_cubic)
            input_wpt = input_path.new_waypoint("input_wpt", character.input_coord)
            explode_scn = character.animation.new_scene("explode")
            explode_star_color = random.choice(star_colors)
            explode_scn.add_frame(character.input_symbol, 1, color=explode_star_color)
            cooling_scn = character.animation.new_scene("cooling")
            cooling_gradient = graphics.Gradient([explode_star_color, self.args.final_color], 10)
            for color in cooling_gradient:
                cooling_scn.add_frame(character.input_symbol, 20, color=color)
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, nearby_path, EventHandler.Action.ACTIVATE_PATH, input_path
            )
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, nearby_path, EventHandler.Action.ACTIVATE_SCENE, cooling_scn
            )
            character.animation.activate_scene(explode_scn)
            character.motion.activate_path(nearby_path)
            self.animating_chars.append(character)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by creating the starfield, blackhole, and consumption scenes/waypoints."""
        self.prepare_blackhole()

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        formation_delay = max(100 // len(self.blackhole_chars), 10)
        f_delay = formation_delay
        next_char_consuming_delay = 0
        max_consume = max(min(int(0.1 * len(self.terminal.characters)), 15), 2)
        phase = "forming"
        awaiting_blackhole_chars = list(self.blackhole_chars)
        while self.animating_chars or phase != "complete":
            if phase == "forming":
                if awaiting_blackhole_chars:
                    if not f_delay:
                        next_char = awaiting_blackhole_chars.pop(0)
                        next_char.motion.activate_path(next_char.motion.paths["blackhole"])
                        next_char.animation.activate_scene(next_char.animation.scenes["blackhole"])
                        self.animating_chars.append(next_char)
                        f_delay = formation_delay
                    else:
                        f_delay -= 1
                else:
                    if not self.animating_chars:
                        self.rotate_blackhole()
                        phase = "consuming"
            elif phase == "consuming":
                if self.awaiting_consumption_chars:
                    if not next_char_consuming_delay:
                        for _ in range(random.randrange(1, max_consume)):
                            if self.awaiting_consumption_chars:
                                next_char = self.awaiting_consumption_chars.pop(0)
                                next_char.motion.activate_path(next_char.motion.paths["singularity"])
                                self.animating_chars.append(next_char)
                            else:
                                break
                        max_consume += 1
                        next_char_consuming_delay = random.randrange(0, 10)
                    else:
                        next_char_consuming_delay -= 1

                else:
                    if all(character in self.blackhole_chars for character in self.animating_chars):
                        phase = "collapsing"

            elif phase == "collapsing":
                self.collapse_blackhole()
                phase = "exploding"
            elif phase == "exploding":
                if all(
                    [
                        character.motion.active_path is None and character.animation.active_scene is None
                        for character in self.blackhole_chars
                    ]
                ):
                    self.explode_singularity()
                    phase = "complete"

            self.animate_chars()
            self.terminal.print()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()



import argparse
import random

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter
from terminaltexteffects.base_character import EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes, motion, easing


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "bubbles",
        formatter_class=argtypes.CustomFormatter,
        help="Characters are formed into bubbles that float down and pop.",
        description="Characters are formed into bubbles that float down and pop.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects bubbles -a 0.01 --pop-color ff9600 --final-color 252525 --bubble-speed 0.1 --bubble-delay 50 --easing IN_OUT_SINE""",
    )
    effect_parser.set_defaults(effect_class=BubblesEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--no-rainbow",
        action="store_true",
        help="If set, the bubbles will not be colored with a rainbow gradient.",
    )
    effect_parser.add_argument(
        "--bubble-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the bubbles. Ignored if --no-rainbow is left as default False.",
    )
    effect_parser.add_argument(
        "--pop-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the spray emitted when a bubble pops.",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="ffffff",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character.",
    )
    effect_parser.add_argument(
        "--bubble-speed",
        type=argtypes.valid_speed,
        default=0.1,
        metavar="(float > 0)",
        help="Speed of the floating bubbles. Note: Speed effects the number of steps in the easing function. Adjust speed and animation rate separately to fine tune the effect.",
    )
    effect_parser.add_argument(
        "--bubble-delay",
        type=argtypes.positive_int,
        default=50,
        metavar="(int > 0)",
        help="Number of animation steps between bubbles.",
    )
    effect_parser.add_argument(
        "--pop-condition",
        default="row",
        choices=["row", "bottom", "anywhere"],
        help="Condition for a bubble to pop. 'row' will pop the bubble when it reaches the the lowest row for which a character in the bubble originates. 'bottom' will pop the bubble at the bottom row of the terminal. 'anywhere' will pop the bubble randomly, or at the bottom of the terminal.",
    )
    effect_parser.add_argument(
        "--easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for character movement after a bubble pops.",
    )


class Bubble:
    def __init__(
        self,
        effect: "BubblesEffect",
        origin: motion.Coord,
        characters: list[EffectCharacter],
    ):
        self.effect = effect
        self.characters = characters
        self.radius = max(len(self.characters) // 5, 1)
        self.origin = origin
        self.anchor_char = EffectCharacter(" ", origin.column, origin.row)
        if self.effect.args.pop_condition == "row":
            self.lowest_row = min([char.input_coord.row for char in self.characters])
        else:
            self.lowest_row = self.effect.terminal.output_area.bottom
        self.set_character_coordinates()
        self.landed = False
        self.make_waypoints()
        self.make_gradients()

    def set_character_coordinates(self) -> None:
        for i, char in enumerate(self.characters):
            point = self.anchor_char.motion.find_coords_on_circle(
                self.anchor_char.motion.current_coord,
                self.radius,
                len(self.characters),
            )[i]
            char.motion.set_coordinate(point)
            if point.row == self.lowest_row:
                self.landed = True

        if self.effect.args.pop_condition == "anywhere":
            if random.random() < 0.002:
                self.landed = True

    def make_waypoints(self):
        waypoint_column = random.randint(self.effect.terminal.output_area.left, self.effect.terminal.output_area.right)
        floor_path = self.anchor_char.motion.new_path("floor", speed=self.effect.args.bubble_speed)
        floor_waypoint = floor_path.new_waypoint("floor", motion.Coord(waypoint_column, self.lowest_row))
        self.anchor_char.motion.activate_path(floor_path)

    def make_gradients(self) -> None:
        if self.effect.args.no_rainbow:
            for character in self.characters:
                sheen_scene = character.animation.new_scene("sheen")
                sheen_scene.add_frame(character.input_symbol, 1, color=self.effect.args.bubble_color)
                character.animation.activate_scene(sheen_scene)
        else:
            rainbow_gradient = list(self.effect.rainbow_gradient.spectrum)
            gradient_offset = 0
            for character in self.characters:
                sheen_scene = character.animation.new_scene("sheen")
                for step in rainbow_gradient:
                    sheen_scene.add_frame(character.input_symbol, 5, color=step)
                gradient_offset += 2
                gradient_offset %= len(rainbow_gradient)
                rainbow_gradient = rainbow_gradient[gradient_offset:] + rainbow_gradient[:gradient_offset]
                character.animation.activate_scene(sheen_scene)
                if character.animation.active_scene:
                    character.animation.active_scene.is_looping = True

    def pop(self) -> None:
        char: EffectCharacter
        point: motion.Coord
        for char, point in zip(
            self.characters,
            self.anchor_char.motion.find_coords_on_circle(
                self.anchor_char.motion.current_coord,
                self.radius + 3,
                len(self.characters),
            ),
        ):
            pop_out_path = char.motion.new_path("pop_out", speed=0.2, ease=easing.out_expo)
            pop_out_waypoint = pop_out_path.new_waypoint("pop_out", point)
            char.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE,
                pop_out_path,
                EventHandler.Action.ACTIVATE_PATH,
                char.motion.paths["final"],
            )
        for character in self.characters:
            character.animation.activate_scene(character.animation.scenes["pop_1"])
            character.motion.activate_path(character.motion.paths["pop_out"])

    def activate(self) -> None:
        for char in self.characters:
            char.is_active = True

    def move(self) -> None:
        self.anchor_char.motion.move()
        self.set_character_coordinates()
        for character in self.characters:
            character.animation.step_animation()


class BubblesEffect:
    """Effect that ___."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_chars: list[EffectCharacter] = []
        self.animating_chars: list[EffectCharacter] = []
        self.bubbles: list[Bubble] = []
        red = "e81416"
        orange = "ffa500"
        yellow = "faeb36"
        green = "79c314"
        blue = "487de7"
        indigo = "4b369d"
        violet = "70369d"
        self.rainbow_gradient = graphics.Gradient([red, orange, yellow, green, blue, indigo, violet], 5)

    def prepare_data(self) -> None:
        """Prepares the data for the effect by ___."""
        final_gradient = graphics.Gradient([self.args.pop_color, self.args.final_color], 10)
        for character in self.terminal.characters:
            character.layer = 1
            pop_1_scene = character.animation.new_scene("pop_1")
            pop_2_scene = character.animation.new_scene("pop_2")
            pop_1_scene.add_frame("*", 25, color=self.args.pop_color)
            pop_2_scene.add_frame("'", 25, color=self.args.pop_color)
            final_scene = character.animation.new_scene("final")
            for step in final_gradient:
                final_scene.add_frame(character.input_symbol, 5, color=step)
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, pop_1_scene, EventHandler.Action.ACTIVATE_SCENE, pop_2_scene
            )
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE,
                pop_2_scene,
                EventHandler.Action.ACTIVATE_SCENE,
                final_scene,
            )
            final_path = character.motion.new_path(
                "final",
                speed=0.3,
                ease=easing.in_out_expo,
            )
            final_waypoint = final_path.new_waypoint("final", character.input_coord)
            character.event_handler.register_event(
                EventHandler.Event.PATH_COMPLETE, final_path, EventHandler.Action.SET_LAYER, 0
            )

        unbubbled_chars = []
        for _, char_list in self.terminal.input_by_row().items():
            unbubbled_chars.extend(char_list)
        self.bubbles = []
        while unbubbled_chars:
            bubble_group = []
            if len(unbubbled_chars) < 5:
                bubble_group.extend(unbubbled_chars)
                unbubbled_chars.clear()
            else:
                for _ in range(random.randint(5, min(len(unbubbled_chars), 20))):
                    bubble_group.append(unbubbled_chars.pop(0))
            bubble_origin = motion.Coord(
                random.randint(self.terminal.output_area.left, self.terminal.output_area.right),
                self.terminal.output_area.top,
            )
            new_bubble = Bubble(self, bubble_origin, bubble_group)
            self.bubbles.append(new_bubble)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        animating_bubbles: list[Bubble] = []
        steps_since_last_bubble = 0
        while animating_bubbles or self.animating_chars or self.bubbles:
            if self.bubbles and steps_since_last_bubble >= self.args.bubble_delay:
                next_bubble = self.bubbles.pop(0)
                next_bubble.activate()
                animating_bubbles.append(next_bubble)
                steps_since_last_bubble = 0
            steps_since_last_bubble += 1

            for bubble in animating_bubbles:
                if bubble.landed:
                    bubble.pop()
                    self.animating_chars.extend(bubble.characters)

            animating_bubbles = [bubble for bubble in animating_bubbles if not bubble.landed]
            self.animate_bubbles(animating_bubbles)
            self.animate_chars()
            self.terminal.print()

            # remove completed chars from animating chars
            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_bubbles(self, animating_bubbles: list[Bubble]) -> None:
        for bubble in animating_bubbles:
            bubble.move()

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation."""
        for animating_char in self.animating_chars:
            animating_char.animation.step_animation()
            animating_char.motion.move()



"""Classes for storing and manipulating character graphics."""

import typing
import random
from enum import Enum, auto
from dataclasses import dataclass
import itertools

from terminaltexteffects.utils import ansitools, colorterm, hexterm, motion, easing

if typing.TYPE_CHECKING:
    from terminaltexteffects import base_character


class SyncMetric(Enum):
    """Enum for specifying the type of sync to use for a Scene.

    Attributes:
        DISTANCE (int): Sync to a Waypoint based on distance from the Waypoint
        STEP (int): Sync to a Waypoint based on the number of steps taken towards the Waypoint
    """

    DISTANCE = auto()
    STEP = auto()


@dataclass
class Gradient:
    """A Gradient is a list of RGB hex color strings transitioning from one color to another. The gradient color
    list is calculated using linear interpolation based on the provided start and end colors and the number of steps. Gradients
    can be interated over to get the next color in the gradient color list.

    Args:
        stops list[str]: List of RGB hex color string or XTerm-256 color codes. Each stop will have steps number of frames between it and the next stop.
        steps (int): Number of intermediate colors to calculate

    Attributes:
        colors (list[str]): List (length=steps + 2) of RGB hex color strings

    """

    stops: list[int | str]
    steps: int

    def __post_init__(self) -> None:
        self.spectrum: list[str] = self._generate()
        self.index: int = 0

    def _generate(self) -> list[str]:
        """Calculate a gradient of colors between two colors using linear interpolation.

        Returns:
            list[str]: List (length=steps + 2) of RGB hex color strings
        """
        if len(self.stops) < 2:
            raise ValueError("Gradient must have at least 2 stops.")
        spectrum: list[str] = []
        color_pairs = itertools.pairwise(self.stops)
        for start, end in color_pairs:
            # Convert start_color to hex if it's an XTerm-256 color code
            if isinstance(start, int):
                start = hexterm.xterm_to_hex(start)
            # Convert end_color to hex if it's an XTerm-256 color code
            if isinstance(end, int):
                end = hexterm.xterm_to_hex(end)
            # Convert start_color to a list of RGB values
            start_color_ints = colorterm._hex_to_int(start)
            # Convert end_color to a list of RGB values
            end_color_ints = colorterm._hex_to_int(end)
            # Initialize an empty list to store the gradient colors
            gradient_colors: list[str] = []
            # Calculate the color deltas for each RGB value
            red_delta = (end_color_ints[0] - start_color_ints[0]) // self.steps
            green_delta = (end_color_ints[1] - start_color_ints[1]) // self.steps
            blue_delta = (end_color_ints[2] - start_color_ints[2]) // self.steps
            # Calculate the intermediate colors and add them to the gradient colors list
            for i in range(max(self.steps - 1, 0)):
                red = start_color_ints[0] + (red_delta * i)
                green = start_color_ints[1] + (green_delta * i)
                blue = start_color_ints[2] + (blue_delta * i)

                # Ensure that the RGB values are within the valid range of 0-255
                red = max(0, min(red, 255))
                green = max(0, min(green, 255))
                blue = max(0, min(blue, 255))

                # Convert the RGB values to a hex color string and add it to the gradient colors list
                gradient_colors.append(f"{red:02x}{green:02x}{blue:02x}")
            # Add the end color to the gradient colors list
            gradient_colors.append(end)
            spectrum.extend(gradient_colors)
        return spectrum

    def __iter__(self) -> "Gradient":
        self.index = 0
        return self

    def __next__(self) -> str:
        if self.index < len(self.spectrum):
            color = self.spectrum[self.index]
            self.index += 1
            return color
        else:
            raise StopIteration

    def __add__(self, other: "Gradient") -> "Gradient":
        if not isinstance(other, Gradient):
            return NotImplemented
        new_gradient = Gradient(["ffffff", "ffffff"], 1)
        new_gradient.spectrum = self.spectrum + other.spectrum
        new_gradient.steps = self.steps + other.steps
        new_gradient.stops = self.stops + other.stops
        return new_gradient


@dataclass
class CharacterVisual:
    """A class for storing symbol, color, and terminal graphical modes for the character.

    Supported graphical modes:
    bold, dim, italic, underline, blink, inverse, hidden, strike

    Args:
        symbol (str): the symbol to show
        bold (bool): bold mode
        dim (bool): dim mode
        italic (bool): italic mode
        underline (bool): underline mode
        blink (bool): blink mode
        reverse (bool): reverse mode
        hidden (bool): hidden mode
        strike (bool): strike mode
        color (int | str): color code
    """

    symbol: str
    bold: bool = False
    dim: bool = False
    italic: bool = False
    underline: bool = False
    blink: bool = False
    reverse: bool = False
    hidden: bool = False
    strike: bool = False
    color: int | str | None = None

    def __post_init__(self):
        self.format_symbol()

    def disable_modes(self) -> None:
        """Disables all graphical modes."""
        self.bold = False
        self.dim = False
        self.italic = False
        self.underline = False
        self.blink = False
        self.reverse = False
        self.hidden = False
        self.strike = False

    def format_symbol(self) -> None:
        """Formats the symbol for printing."""
        formatting_string = ""
        if self.bold:
            formatting_string += ansitools.APPLY_BOLD()
        if self.italic:
            formatting_string += ansitools.APPLY_ITALIC()
        if self.underline:
            formatting_string += ansitools.APPLY_UNDERLINE()
        if self.blink:
            formatting_string += ansitools.APPLY_BLINK()
        if self.reverse:
            formatting_string += ansitools.APPLY_REVERSE()
        if self.hidden:
            formatting_string += ansitools.APPLY_HIDDEN()
        if self.strike:
            formatting_string += ansitools.APPLY_STRIKETHROUGH()
        if self.color is not None:
            formatting_string += colorterm.fg(self.color)

        self.symbol = f"{formatting_string}{self.symbol}{ansitools.RESET_ALL() if formatting_string else ''}"


@dataclass
class Frame:
    """A Frame is a CharacterVisual with a duration.
    Args:
        character_visual (CharacterVisual): a CharacterVisual object
        duration (int): the number of animation steps to use the Frame
    """

    character_visual: CharacterVisual
    duration: int

    def __post_init__(self):
        self.frames_played = 0
        self.symbol = self.character_visual.symbol


class Scene:
    """A Scene is a list of Frames.

    Args:
        scene_id (str): unique ID for the Scene
        is_looping (bool): whether the Scene should loop
        sync (SyncMetric): the sync metric to use
        no_color (bool): whether to disable color
        use_xterm_colors (bool): whether to use XTerm-256 colors
    """

    xterm_color_map: dict[str, int] = {}

    def __init__(
        self,
        scene_id: str,
        is_looping: bool = False,
        sync: SyncMetric | None = None,
        ease: typing.Callable | None = None,
        no_color: bool = False,
        use_xterm_colors: bool = False,
    ):
        self.scene_id = scene_id
        self.is_looping = is_looping
        self.sync: SyncMetric | None = sync
        self.ease: easing.Ease | None = ease
        self.no_color = no_color
        self.use_xterm_colors = use_xterm_colors
        self.frames: list[Frame] = []
        self.played_frames: list[Frame] = []
        self.frame_index_map: dict[int, Frame] = {}
        self.easing_total_steps: int = 0
        self.easing_current_step: int = 0

    def add_frame(
        self,
        symbol: str,
        duration: int,
        *,
        color: str | int | None = None,
        bold=False,
        dim=False,
        italic=False,
        underline=False,
        blink=False,
        reverse=False,
        hidden=False,
        strike=False,
    ) -> None:
        """Adds a Frame to the Scene.

        Args:
            symbol (str): the symbol to show
            color (str | int): color code
            duration (int): the number of animation steps to use the Frame
            BOLD (bool): bold mode
            DIM (bool): dim mode
            ITALIC (bool): italic mode
            UNDERLINE (bool): underline mode
            BLINK (bool): blink mode
            REVERSE (bool): reverse mode
            HIDDEN (bool): hidden mode
            STRIKE (bool): strike mode
        """
        if not self.no_color and color is not None:
            if self.use_xterm_colors and isinstance(color, str):
                if color in self.xterm_color_map:
                    color = self.xterm_color_map[color]
                else:
                    xterm_color = hexterm.hex_to_xterm(color)
                    self.xterm_color_map[color] = xterm_color
                    color = xterm_color
        if duration < 1:
            raise ValueError("duration must be greater than 0")
        char_vis = CharacterVisual(
            symbol,
            bold=bold,
            dim=dim,
            italic=italic,
            underline=underline,
            blink=blink,
            reverse=reverse,
            hidden=hidden,
            strike=strike,
            color=color,
        )
        frame = Frame(char_vis, duration)
        self.frames.append(frame)
        for _ in range(frame.duration):
            self.frame_index_map[self.easing_total_steps] = frame
            self.easing_total_steps += 1

    def activate(self) -> str:
        """Activates the Scene.

        Returns:
            str: the next symbol in the Scene
        """
        if self.frames:
            return self.frames[0].symbol
        else:
            raise ValueError("Scene has no sequences.")

    def apply_gradient(self, start_color: str | int, end_color: str | int) -> None:
        """
        Applies a gradient effect across the frames of the scene. The gradient is generated
        to have the same number of steps as the number of frames in the scene.

        Parameters
        ----------
        start_color : str | int
            The starting color of the gradient.
        end_color : str | int
            The ending color of the gradient.

        Returns
        -------
        None
        """
        if not self.frames:
            return
        else:
            gradient = Gradient([start_color, end_color], len(self.frames))
            for i, frame in enumerate(self.frames):
                frame.character_visual.color = gradient.spectrum[i]
                frame.character_visual.format_symbol()

    def get_next_symbol(self) -> str:
        """Returns the next symbol in the Scene.

        Returns:
            str: the next symbol in the Scene
        """

        current_sequence = self.frames[0]
        next_symbol = current_sequence.symbol
        current_sequence.frames_played += 1
        if current_sequence.frames_played == current_sequence.duration:
            current_sequence.frames_played = 0
            self.played_frames.append(self.frames.pop(0))
            if self.is_looping and not self.frames:
                self.frames.extend(self.played_frames)
                self.played_frames.clear()
        return next_symbol

    def reset_scene(self) -> None:
        """Resets the Scene."""
        for sequence in self.frames:
            sequence.frames_played = 0
            self.played_frames.append(sequence)
        self.frames.clear()
        self.frames.extend(self.played_frames)
        self.played_frames.clear()

    def __eq__(self, other: typing.Any):
        if not isinstance(other, Scene):
            return NotImplemented
        return self.scene_id == other.scene_id

    def __hash__(self):
        return hash(self.scene_id)


class Animation:
    def __init__(self, character: "base_character.EffectCharacter"):
        """Animation handles the animations of a character. It contains a scene_name -> Scene mapping and the active Scene. Calls to step_animation()
        progress the Scene and apply the next symbol to the character.

        Args:
            character (base_character.EffectCharacter): the EffectCharacter object to animate
        """
        self.scenes: dict[str, Scene] = {}
        self.character = character
        self.active_scene: Scene | None = None
        self.use_xterm_colors: bool = False
        self.no_color: bool = False
        self.xterm_color_map: dict[str, int] = {}
        self.active_scene_current_step: int = 0

    def new_scene(
        self,
        scene_name: str,
        *,
        is_looping: bool = False,
        sync: SyncMetric | None = None,
        ease: typing.Callable | None = None,
    ) -> Scene:
        """Creates a new Scene and adds it to the Animation.

        Args:
            scene_name (str): Unique name for the scene. Used to reference the scene outside the scope in which is was created.
            is_looping (bool): Whether the scene should loop.
            sync (SyncMetric): The type of sync to use for the scene.
            ease (typing.Callable): The easing function to use for the scene.

        Returns:
            Scene: the new Scene
        """
        new_scene = Scene(scene_name, is_looping=is_looping, sync=sync, ease=ease)
        self.scenes[scene_name] = new_scene
        new_scene.no_color = self.no_color
        new_scene.use_xterm_colors = self.use_xterm_colors
        return new_scene

    def active_scene_is_complete(self) -> bool:
        """Returns whether the active scene is complete. A scene is complete if all sequences have been played.
        Looping scenes are always complete.

        Returns:
            bool: True if complete, False otherwise
        """
        if not self.active_scene:
            return True
        elif not self.active_scene.frames or self.active_scene.is_looping:
            return True

        return False

    @staticmethod
    def random_color() -> str:
        """Returns a random hex color code.

        Returns:
            str: hex color code
        """
        return hex(random.randint(0, 0xFFFFFF))[2:].zfill(6)

    def _ease_animation(self, easing_func: typing.Callable) -> float:
        """Returns the percentage of total distance that should be moved based on the easing function.

        Args:
            easing_func (Callable): The easing function to use.

        Returns:
            float: The percentage of total distance to move.
        """

        if self.active_scene is None:
            return 0
        elapsed_step_ratio = self.active_scene.easing_current_step / self.active_scene.easing_total_steps
        return easing_func(elapsed_step_ratio)

    def step_animation(self) -> None:
        """Apply the next symbol in the scene to the character. If a scene order exists, the next scene
        will be activated when the current scene is complete."""
        if self.active_scene and self.active_scene.frames:
            # if the active scene is synced to movement, calculate the sequence index based on the
            # current waypoint progress
            if self.active_scene.sync:
                if self.character.motion.active_path:
                    if self.active_scene.sync == SyncMetric.STEP:
                        sequence_index = round(
                            (len(self.active_scene.frames) - 1)
                            * (
                                max(self.character.motion.active_path.current_step, 1)
                                / max(self.character.motion.active_path.max_steps, 1)
                            )
                        )
                    elif self.active_scene.sync == SyncMetric.DISTANCE:
                        sequence_index = round(
                            (len(self.active_scene.frames) - 1)
                            * (
                                max(
                                    max(self.character.motion.active_path.total_distance, 1)
                                    - max(
                                        self.character.motion.active_path.total_distance
                                        - self.character.motion.active_path.last_distance_reached,
                                        1,
                                    ),
                                    1,
                                )
                                / max(self.character.motion.active_path.total_distance, 1)
                            )
                        )
                    try:
                        self.character.symbol = self.active_scene.frames[sequence_index].symbol
                    except IndexError:
                        self.character.symbol = self.active_scene.frames[-1].symbol
                else:  # when the active waypoint has been deactivated, use the final symbol in the scene and finish the scene
                    self.character.symbol = self.active_scene.frames[-1].symbol
                    self.active_scene.played_frames.extend(self.active_scene.frames)
                    self.active_scene.frames.clear()

            elif self.active_scene and self.active_scene.ease:
                easing_factor = self._ease_animation(self.active_scene.ease)
                frame_index = round(easing_factor * max(self.active_scene.easing_total_steps - 1, 0))
                frame_index = max(min(frame_index, self.active_scene.easing_total_steps - 1), 0)
                frame = self.active_scene.frame_index_map[frame_index]
                self.character.symbol = frame.symbol
                self.active_scene.easing_current_step += 1
                if self.active_scene.easing_current_step == self.active_scene.easing_total_steps:
                    if self.active_scene.is_looping:
                        self.active_scene.easing_current_step = 0
                    else:
                        self.active_scene.played_frames.extend(self.active_scene.frames)
                        self.active_scene.frames.clear()

            else:
                self.character.symbol = self.active_scene.get_next_symbol()
            if self.active_scene_is_complete():
                completed_scene = self.active_scene
                if not self.active_scene.is_looping:
                    self.active_scene.reset_scene()
                    self.active_scene = None

                self.character.event_handler.handle_event(
                    self.character.event_handler.Event.SCENE_COMPLETE, completed_scene
                )

    def activate_scene(self, scene: Scene) -> None:
        """Sets the active scene.

        Args:
            scene (Scene): the Scene to set as active
        """
        self.active_scene = scene
        self.active_scene_current_step = 0
        self.character.symbol = self.active_scene.activate()
        self.character.event_handler.handle_event(self.character.event_handler.Event.SCENE_ACTIVATED, scene)

    def deactivate_scene(self, scene: Scene) -> None:
        """Deactivates a scene.

        Args:
            scene (Scene): the Scene to deactivate
        """
        if self.active_scene is scene:
            self.active_scene = None



import argparse

import terminaltexteffects.utils.argtypes as argtypes
from terminaltexteffects.base_character import EffectCharacter, EventHandler
from terminaltexteffects.utils.terminal import Terminal
from terminaltexteffects.utils import graphics, argtypes


def add_arguments(subparsers: argparse._SubParsersAction) -> None:
    """Adds arguments to the subparser.

    Args:
        subparser (argparse._SubParsersAction): subparser to add arguments to
    """
    effect_parser = subparsers.add_parser(
        "waves",
        formatter_class=argtypes.CustomFormatter,
        help="Waves travel across the terminal leaving behind the characters.",
        description="Waves travel across the terminal leaving behind the characters.",
        epilog=f"""{argtypes.EASING_EPILOG}

Example: terminaltexteffects waves -a 0.01 --wave-deep-color 4651e3 --wave-shallow-color 57FB92 --final-color 80F7E6 --wave-count 6 --wave-easing IN_OUT_SINE""",
    )
    effect_parser.set_defaults(effect_class=WavesEffect)
    effect_parser.add_argument(
        "-a",
        "--animation-rate",
        type=argtypes.valid_animationrate,
        default=0.01,
        help="Minimum time, in seconds, between animation steps. This value does not normally need to be modified. Use this to increase the playback speed of all aspects of the effect. This will have no impact beyond a certain lower threshold due to the processing speed of your device.",
    )
    effect_parser.add_argument(
        "--wave-deep-color",
        type=argtypes.valid_color,
        default="4651e3",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the deepest part of the wave. A gradient is generated between the shallow and deep parts of the wave.",
    )
    effect_parser.add_argument(
        "--wave-shallow-color",  # make more descriptive
        type=argtypes.valid_color,
        default="57FB92",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the shallow part of the wave. A gradient is generated between the shallow and deep parts of the wave..",
    )
    effect_parser.add_argument(
        "--final-color",
        type=argtypes.valid_color,
        default="80F7E6",
        metavar="(XTerm [0-255] OR RGB Hex [000000-ffffff])",
        help="Color for the final character. A gradient is generated between the shallow part of the wave and the final color.",
    )
    effect_parser.add_argument(
        "--wave-count",
        type=argtypes.positive_int,
        default=6,
        help="Number of waves to generate. n > 0.",
    )
    effect_parser.add_argument(
        "--wave-easing",
        default="IN_OUT_SINE",
        type=argtypes.valid_ease,
        help="Easing function to use for wave travel.",
    )


class WavesEffect:
    """Effect that creates waves that travel across the terminal, leaving behind the characters."""

    def __init__(self, terminal: Terminal, args: argparse.Namespace):
        self.terminal = terminal
        self.args = args
        self.pending_columns: list[list[EffectCharacter]] = []
        self.animating_chars: list[EffectCharacter] = []

    def prepare_data(self) -> None:
        """Prepares the data for the effect by creating the wave animations."""
        block_wipe_start = ("▁", "▂", "▃", "▄", "▅", "▆", "▇", "█")
        block_wipe_end = ("▇", "▆", "▅", "▄", "▃", "▂", "▁")
        wave_gradient_light = graphics.Gradient([self.args.wave_shallow_color, self.args.wave_deep_color], 8)
        wave_gradient_dark = graphics.Gradient([self.args.wave_deep_color, self.args.wave_shallow_color], 8)
        wave_gradient = list(wave_gradient_light) + list(wave_gradient_dark)[1:]
        colored_waves = list(zip(block_wipe_start + block_wipe_end, wave_gradient))
        final_gradient = graphics.Gradient([self.args.wave_shallow_color, self.args.final_color], 8)
        for character in self.terminal.characters:
            wave_scn = character.animation.new_scene("wave")
            wave_scn.ease = self.args.wave_easing
            for _ in range(self.args.wave_count):
                for block, color in colored_waves:
                    wave_scn.add_frame(block, 3, color=color)
            final_scn = character.animation.new_scene("final")
            for step in final_gradient:
                final_scn.add_frame(character.input_symbol, 15, color=step)
            character.event_handler.register_event(
                EventHandler.Event.SCENE_COMPLETE, wave_scn, EventHandler.Action.ACTIVATE_SCENE, final_scn
            )
            character.animation.activate_scene(wave_scn)
        columns = self.terminal.input_by_column()
        for _, column in columns.items():
            self.pending_columns.append(column)

    def run(self) -> None:
        """Runs the effect."""
        self.prepare_data()
        while self.pending_columns or self.animating_chars:
            if self.pending_columns:
                next_column = self.pending_columns.pop(0)
                for character in next_column:
                    character.is_active = True
                    self.animating_chars.append(character)
            self.terminal.print()
            self.animate_chars()

            self.animating_chars = [
                animating_char
                for animating_char in self.animating_chars
                if not animating_char.animation.active_scene_is_complete()
                or not animating_char.motion.movement_is_complete()
            ]

    def animate_chars(self) -> None:
        """Animates the characters by calling the move method and step animation. Move characters prior to stepping animation
        to ensure waypoint synced animations have the latest waypoint progress information."""
        for animating_char in self.animating_chars:
            animating_char.motion.move()
            animating_char.animation.step_animation()
